﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
        GO_TO_JEWEL,
		SACK_JEWEL,
		EAT_FOOD
    }

    public class ClarionAgent
    {
        #region Constants

		public List<Thing> knapsack = new List<Thing>();

		/// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one jewel close
        /// </summary>
		private String DIMENSION_JEWEL_CLOSE = "Jewel_Close";
		private String DIMENSION_JEWEL_SACK = "Jewel_Sack";
		private String DIMENSION_FOOD_EAT = "Food_Eat";
		private String DIMENSION_DONE = "Done";

        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a jewel on vision.
        /// </summary>
		private DimensionValuePair InputJewelClose;

		private DimensionValuePair InputJewelSack;

		private DimensionValuePair InputFoodEat;

		private DimensionValuePair InputDone;

        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        /// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
		private ExternalActionChunk OutputGoToJewel;
		private ExternalActionChunk OutputSackJewel;
		private ExternalActionChunk OutputEatFood;
		private ExternalActionChunk OutputDone;

		public IList<Thing> CurrentSensorialInfo;

        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputJewelClose = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_JEWEL_CLOSE);

			InputJewelSack = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_JEWEL_SACK);

			InputFoodEat = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_FOOD_EAT);

			InputDone = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_DONE);

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
            OutputGoToJewel = World.NewExternalActionChunk(ClarionAgentActionType.GO_TO_JEWEL.ToString());
			OutputSackJewel = World.NewExternalActionChunk(ClarionAgentActionType.SACK_JEWEL.ToString());
			OutputEatFood = World.NewExternalActionChunk(ClarionAgentActionType.EAT_FOOD.ToString ());
			OutputDone = World.NewExternalActionChunk (ClarionAgentActionType.DO_NOTHING.ToString ());

            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {
     
			SupportCalculator lookForJewelSupportCalculator = FixedRuleDelegateToLookForJewel;
			FixedRule ruleLookForJewel = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotateClockwise, lookForJewelSupportCalculator);
			CurrentAgent.Commit(ruleLookForJewel);

            SupportCalculator goToJewelSupportCalculator = FixedRuleDelegateToGoToJewel;
			FixedRule ruleToGoToJewel = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoToJewel, goToJewelSupportCalculator);
            CurrentAgent.Commit(ruleToGoToJewel);

			SupportCalculator sackJewelSupportCalculator = FixedRuleDelegateToSackJewel;
			FixedRule ruleToSackJewel = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputSackJewel, sackJewelSupportCalculator);
			CurrentAgent.Commit(ruleToSackJewel);

			SupportCalculator eatFoodSupportCalculator = FixedRuleDelegateToEatFood;
			FixedRule ruleToEatFood = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputEatFood, eatFoodSupportCalculator);
			CurrentAgent.Commit(ruleToEatFood);

			SupportCalculator doneSupportCalculator = FixedRuleDelegateDone;
			FixedRule ruleDone = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, OutputDone, doneSupportCalculator);
			CurrentAgent.Commit (ruleDone);

            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
			
            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);

            // Detect if we have a wall ahead

			Boolean hasJewel = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_JEWEL)).Any();
			double hasJewelActivationValue = hasJewel ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;

			Boolean hasFood = sensorialInformation.Where(item => (item.CategoryId == Thing.categoryPFOOD || item.CategoryId == Thing.CATEGORY_NPFOOD)).Any();

			Thing closestFood = null, closestJewel = null;
			double eatFoodActivationValue = CurrentAgent.Parameters.MIN_ACTIVATION;
			double sackJewelActivationValue = CurrentAgent.Parameters.MIN_ACTIVATION;

			if (hasFood) {
				closestFood = sensorialInformation.Where (item => (item.CategoryId == Thing.categoryPFOOD || item.CategoryId == Thing.CATEGORY_NPFOOD)).OrderBy (item => item.DistanceToCreature).First ();
				if (closestFood.DistanceToCreature <= 40) {
					eatFoodActivationValue = CurrentAgent.Parameters.MAX_ACTIVATION;
				}
			}
			
			if (hasJewel) {
				closestJewel = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_JEWEL)).OrderBy (item => item.DistanceToCreature).First ();
				if (closestJewel.DistanceToCreature <= 30) {
					sackJewelActivationValue = CurrentAgent.Parameters.MAX_ACTIVATION;

				}
			}

			if (eatFoodActivationValue == CurrentAgent.Parameters.MAX_ACTIVATION 
				&& sackJewelActivationValue == CurrentAgent.Parameters.MAX_ACTIVATION) {
				//Both Jewel and Food are close.. get the closest one!
				if (closestJewel.DistanceToCreature <= closestFood.DistanceToCreature) {
					eatFoodActivationValue *= 0.5f;
				} else {
					sackJewelActivationValue *= 0.5f;
				}
				hasJewelActivationValue *= 0.5f;
			}

			si.Add (InputFoodEat, eatFoodActivationValue);
			si.Add (InputJewelSack, sackJewelActivationValue);
			si.Add (InputJewelClose, hasJewelActivationValue);

			Creature theCreature = sensorialInformation.First(item => (item.CategoryId == Thing.CATEGORY_CREATURE)) as Creature;
			double doneActivationValue = theCreature.isLeafletCompleted() ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
			si.Add (InputDone, doneActivationValue);

            return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

        private double FixedRuleDelegateToLookForJewel(ActivationCollection currentInput, Rule target)
        {
			if (currentInput.Contains (InputDone, CurrentAgent.Parameters.MAX_ACTIVATION)) {
				return 0.0;
			} else 
				
			return ((currentInput.Contains(InputJewelClose, CurrentAgent.Parameters.MIN_ACTIVATION))) ? 1.0 : 0.0;
        }

        private double FixedRuleDelegateToGoToJewel(ActivationCollection currentInput, Rule target)
        {
            // Here we will make the logic to go to jewel
			return ((currentInput.Contains(InputJewelClose, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }

		private double FixedRuleDelegateToSackJewel(ActivationCollection currentInput, Rule target)
		{
			return ((currentInput.Contains(InputJewelSack, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		private double FixedRuleDelegateToEatFood(ActivationCollection currentInput, Rule target)
		{
			return ((currentInput.Contains(InputFoodEat, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		private double FixedRuleDelegateDone(ActivationCollection currentInput, Rule target) {
			return ((currentInput.Contains(InputDone, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

				//Update the leaflet info.
				Creature theCreature = sensorialInformation.First(item => (item.CategoryId == Thing.CATEGORY_CREATURE)) as Creature;
				theCreature.updateLeafletWithKnapsack (knapsack);

				CurrentSensorialInfo = sensorialInformation;

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }

				if (theCreature.isLeafletCompleted ()) {
					Console.WriteLine(" ---- LEAFLET COMPLETED ------ Simulation Halted!");
					CurrentCognitiveCycle = MaxNumberOfCognitiveCycles;
					Abort (true);
				}
            }
        }

        #endregion

        #endregion
    }
}
